@@ -19,7 +19,7 @@ class BuildPy(build_py):
         for package, _, _, filenames in self.data_files:
             for wrapper in self.wrappers:
                 if wrapper.package_name == package:
-                    filenames.extend(wrapper.generated_files)
+                    filenames.extend(wrapper.additional_data_files)
                     break
 
         build_py.run(self)
