@@ -79,6 +79,7 @@ class Hooks:
         self.casters = casters
         self.report_only = report_only
         self.has_operators = False
+        self.has_vcheck = False
 
         self.types: typing.Set[str] = set()
         self.class_hierarchy: typing.Dict[str, typing.List[str]] = {}
@@ -145,7 +146,7 @@ class Hooks:
         return name
 
     def _process_doc(
-        self, thing, data, append_prefix=""
+        self, thing, data, append_prefix="", param_remap: typing.Dict[str, str] = {}
     ) -> typing.Optional[typing.List[str]]:
         doc = ""
 
@@ -153,7 +154,15 @@ class Hooks:
             doc = data.doc
         elif "doxygen" in thing:
             doc = thing["doxygen"]
-            doc = sphinxify.process_raw(doc)
+            if param_remap:
+                d = sphinxify.Doc.from_comment(doc)
+                for param in d.params:
+                    new_name = param_remap.get(param.name)
+                    if new_name:
+                        param.name = new_name
+                doc = str(d)
+            else:
+                doc = sphinxify.process_raw(doc)
 
         if data.doc_append is not None:
             doc += f"\n{append_prefix}" + data.doc_append.replace(
@@ -221,6 +230,46 @@ class Hooks:
 
         return param_sig
 
+    def _process_base_param(self, decl_param):
+        params = decl_param.get("params")
+        if params:
+            # recurse
+            params = [self._process_base_param(param) for param in params]
+            return f"{decl_param['param']}<{', '.join(params)}>"
+        else:
+            return decl_param["param"]
+
+    def _make_base_params(
+        self, base_decl_params, pybase_params: typing.Set[str]
+    ) -> str:
+        base_params = [
+            self._process_base_param(decl_param) for decl_param in base_decl_params
+        ]
+
+        for decl_param in base_params:
+            pybase_params.add(decl_param)
+
+        return ", ".join(base_params)
+
+    def _extract_typealias(
+        self,
+        in_ta: typing.List[str],
+        out_ta: typing.List[str],
+        ta_names: typing.Set[str],
+    ):
+        for typealias in in_ta:
+            if typealias.startswith("template"):
+                out_ta.append(typealias)
+            else:
+                teq = typealias.find("=")
+                if teq != -1:
+                    ta_name = typealias[:teq].strip()
+                    out_ta.append(f"using {typealias}")
+                else:
+                    ta_name = typealias.split("::")[-1]
+                    out_ta.append(f"using {ta_name} = {typealias}")
+                ta_names.add(ta_name)
+
     def _enum_hook(self, en, enum_data):
         ename = en.get("name")
         value_prefix = None
@@ -293,6 +342,11 @@ class Hooks:
         data["templates"] = templates
 
         data["type_caster_includes"] = self._get_type_caster_includes()
+        data["has_vcheck"] = self.has_vcheck
+
+        user_typealias = []
+        self._extract_typealias(self.rawdata.typealias, user_typealias, set())
+        data["user_typealias"] = user_typealias
 
     def _function_hook(self, fn, data: FunctionData, internal: bool = False):
         """shared with methods/functions"""
@@ -315,6 +369,8 @@ class Hooks:
         x_temps = []
         x_keepalives = []
 
+        param_remap = {}
+
         x_genlambda = False
         x_lambda_pre: typing.List[str] = []
         x_lambda_post: typing.List[str] = []
@@ -354,14 +410,32 @@ class Hooks:
             if p["name"] == "":
                 p["name"] = "param%s" % i
             p["x_type"] = p.get("enum", p["raw_type"])
-            p["x_callname"] = p["name"]
-            p["x_retname"] = p["name"]
 
-            po = param_override.get(p["name"])
+            if p["pointer"]:
+                p["x_callname"] = p["name"]
+            elif p["reference"]:
+                p["x_callname"] = f"std::forward<decltype({p['name']})>({p['name']})"
+            else:
+                p["x_callname"] = f"std::move({p['name']})"
+            # this is used in trampoline functions
+            # - separate from x_callname because it's just a passthrough
+            p["x_virtual_callname"] = p["x_callname"]
+
+            orig_pname = p["name"]
+            p["x_retname"] = orig_pname
+
+            po = param_override.get(orig_pname)
             if po:
                 p.update(po.dict(exclude_unset=True))
 
-            p["x_pyarg"] = 'py::arg("%(name)s")' % p
+            py_pname = p["name"]
+            if iskeyword(py_pname):
+                py_pname = f"{py_pname}_"
+
+            if orig_pname != py_pname:
+                param_remap[orig_pname] = py_pname
+
+            p["x_pyarg"] = f'py::arg("{py_pname}")'
 
             if "default" in p:
                 default = self._resolve_default(fn, p, p["default"])
@@ -425,6 +499,8 @@ class Hooks:
             ):
                 if p["pointer"]:
                     p["x_callname"] = f"&{p['x_callname']}"
+                else:
+                    p["x_callname"] = p["name"]
                 ptype = "out"
             elif p["array"]:
                 asz = p.get("array_size", 0)
@@ -506,7 +582,7 @@ class Hooks:
         elif fn["constructor"]:
             x_name = "__init__"
 
-        doc_quoted = self._process_doc(fn, data)
+        doc_quoted = self._process_doc(fn, data, param_remap=param_remap)
 
         if data.keepalive is not None:
             x_keepalives = data.keepalive
@@ -528,6 +604,23 @@ class Hooks:
                     f"{fn['name']}: cannot specify ignore_pure for function that isn't pure"
                 )
 
+            if data.trampoline_cpp_code and (not fn["override"] and not fn["virtual"]):
+                raise ValueError(
+                    f"{fn['name']}: cannot specify trampoline_cpp_code for a non-virtual method"
+                )
+
+            if data.virtual_xform and (not fn["override"] and not fn["virtual"]):
+                raise ValueError(
+                    f"{fn['name']}: cannot specify virtual_xform for a non-virtual method"
+                )
+
+            if fn.get("ref_qualifiers", "") == "&&":
+                # pybind11 doesn't support this, user must fix it
+                if not data.ignore_py and not data.cpp_code:
+                    raise ValueError(
+                        f"{fn['name']}: has && ref-qualifier which cannot be directly bound by pybind11, must specify cpp_code or ignore_py"
+                    )
+
         # if "hook" in data:
         #     eval(data["hook"])(fn, data)
 
@@ -613,26 +706,37 @@ class Hooks:
         pybase_params = set()
 
         for base in cls["inherits"]:
-            bqual = class_data.base_qualnames.get(base["class"])
+            bqual = class_data.base_qualnames.get(base["decl_name"])
             if bqual:
-                base["x_qualname"] = bqual
-            elif "::" not in base["decl_name"]:
-                base["x_qualname"] = f'{cls["namespace"]}::{base["decl_name"]}'
+                base["x_class"] = bqual
+                # TODO: sometimes need to add this to pybase_params, but
+                # that would require parsing this more. Seems sufficiently
+                # obscure, going to omit it for now.
+                tp = bqual.find("<")
+                if tp == -1:
+                    base["x_qualname"] = bqual
+                    base["x_params"] = ""
+                else:
+                    base["x_qualname"] = bqual[:tp]
+                    base["x_params"] = bqual[tp + 1 : -1]
             else:
-                base["x_qualname"] = base["decl_name"]
+                if "::" not in base["decl_name"]:
+                    base["x_qualname"] = f'{cls["namespace"]}::{base["decl_name"]}'
+                else:
+                    base["x_qualname"] = base["decl_name"]
+
+                base_decl_params = base.get("decl_params")
+                if base_decl_params:
+                    base["x_params"] = self._make_base_params(
+                        base_decl_params, pybase_params
+                    )
+                    base["x_class"] = f"{base['x_qualname']}<{base['x_params']}>"
+                else:
+                    base["x_params"] = ""
+                    base["x_class"] = base["x_qualname"]
 
             base["x_qualname_"] = base["x_qualname"].translate(self._qualname_trans)
 
-            base_decl_params = base.get("decl_params")
-            if base_decl_params:
-                for decl_param in base_decl_params:
-                    pybase_params.add(decl_param["param"])
-                base["x_params"] = ", ".join(
-                    decl_param["param"] for decl_param in base_decl_params
-                )
-            else:
-                base["x_params"] = ""
-
         ignored_bases = {ib: True for ib in class_data.ignored_bases}
 
         cls["x_inherits"] = [
@@ -725,6 +829,8 @@ class Hooks:
 
         has_constructor = False
         is_polymorphic = class_data.is_polymorphic
+        vcheck_fns = []
+        cls["x_vcheck_fns"] = vcheck_fns
 
         # bad assumption? yep
         if cls["inherits"]:
@@ -755,6 +861,7 @@ class Hooks:
                     continue
 
                 is_private = access == "private"
+                is_virtual = fn["override"] or fn["virtual"]
 
                 # this has to be done even on private functions, because
                 # we do overload detection here
@@ -763,12 +870,29 @@ class Hooks:
                     fn, signature, cls_key, class_data, is_private
                 )
 
-                if not is_private:
-
+                # Have to process private virtual functions too
+                if not is_private or is_virtual:
                     if method_data.ignore:
                         fn["data"] = method_data
                         continue
 
+                    # If the method has cpp_code defined, it must either match the function
+                    # signature of the method, or virtual_xform must be defined with an
+                    # appropriate conversion. If neither of these are true, it will lead
+                    # to difficult to diagnose errors at runtime. We add a static assert
+                    # to try and catch these errors at compile time
+                    need_vcheck = (
+                        (fn["override"] or fn["virtual"])
+                        and method_data.cpp_code
+                        and not method_data.virtual_xform
+                        and not method_data.trampoline_cpp_code
+                        and not cls["final"]
+                        and not class_data.force_no_trampoline
+                    )
+                    if need_vcheck:
+                        vcheck_fns.append(fn)
+                        self.has_vcheck = True
+
                     if operator:
                         self.has_operators = True
                         if method_data.no_release_gil is None:
@@ -842,3 +966,28 @@ class Hooks:
         cls["x_varname"] = "cls_" + cls_name
         cls["x_name"] = self._set_name(cls_name, class_data)
         cls["x_doc_quoted"] = self._process_doc(cls, class_data)
+        cls["x_inline_code"] = class_data.inline_code or ""
+
+        # do logic for extracting user defined typealiases here
+        # - these are at class scope, so they can include template
+        typealias_names = set()
+        x_user_typealias = []
+        self._extract_typealias(class_data.typealias, x_user_typealias, typealias_names)
+
+        # autodetect embedded using directives, but don't override anything
+        # the user specifies
+        # - these are in block scope, so they cannot include templates
+        x_auto_typealias = []
+        for name, using in cls["using"].items():
+            if (
+                using["access"] == "public"
+                and name not in typealias_names
+                and not using["template"]
+                and using["using_type"] == "typealias"
+            ):
+                x_auto_typealias.append(
+                    f"using {name} [[maybe_unused]] = typename {cls['x_qualname']}::{name}"
+                )
+
+        cls["x_user_typealias"] = x_user_typealias
+        cls["x_auto_typealias"] = x_auto_typealias
