@@ -1,5 +1,6 @@
 from rpytest import ft
 import pytest
+import re
 
 #
 # main module
@@ -46,6 +47,60 @@ def test_good_abstract():
     assert m.mustOverrideMe() == 0x3
 
 
+class MyBadPrivateAbstract(ft.PrivateAbstract):
+    pass
+
+
+def test_private_abstract():
+    # it's private, you can't call it
+    assert not hasattr(ft.PrivateAbstract, "_mustOverrideMe")
+
+
+def test_bad_private_abstract():
+    m = MyBadPrivateAbstract()
+
+    with pytest.raises(
+        RuntimeError,
+        match=r".*"
+        + re.escape(
+            'does not override required function "PrivateAbstract::_mustOverrideMe"'
+        ),
+    ):
+        ft.PrivateAbstract.getPrivateOverride(m)
+
+
+class MyGoodPrivateAbstract(ft.PrivateAbstract):
+    def _mustOverrideMe(self):
+        return 0x3
+
+
+def test_good_private_abstract():
+    m = MyGoodPrivateAbstract()
+    assert m._mustOverrideMe() == 0x3
+    assert ft.PrivateAbstract.getPrivateOverride(m) == 0x3
+
+
+#
+# factory.h
+#
+
+
+def test_factory():
+    o = ft.HasFactory(4)
+    assert o.m_x == 5
+
+
+#
+# inline_code.h
+#
+
+
+def test_inline_code():
+    o = ft.InlineCode()
+    assert o.get2() == 2
+    assert o.get4() == 4
+
+
 #
 # static_only.h
 #
